#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sqlite.h>
#include <unistd.h>

#define r_list_sql                                            \
"  select r.name, r.desc                                    " \
"    from recipes as r                                      " \
"                                                           " \
"  except                                                   " \
"                                                           " \
"  select r.name, r.desc                                    " \
"    from recipes as r, ingredients as n, inventory as i    " \
"   where n.type_id not in (select type_id from inventory)  " \
"     and r.name = n.recipe                                 " \
"                                                           " \
"order by r.name;                                           "

static int dict_cb(void* extra, int argc, char** argv, char** columns)
{
    printf("  %0.8X : %s\n", atoi(argv[0]), argv[1]);
    return 0;
}

static int inv_cb(void* extra, int argc, char** argv, char** columns)
{
    printf("% 3d : %s\n", atoi(argv[1]), argv[0]);
    return 0;
}

static int rlist_cb(void* extra, int argc, char** argv, char** columns)
{
    printf("  %-25s : %s\n", argv[0], argv[1] == NULL ? "" : argv[1]);
    return 0;
}

int main(int argc, char** argv)
{
    char* line = NULL;
    int line_len, line_cap = 0; 
    int i;
    char* cmd, *args;
    char* db_name;
    sqlite* db;
    char* err;

    /* Take the database filename from the first argument. */
    if (argc == 1) db_name = "smartshelf.db";
    else           db_name = argv[1];

    /* Open database. */
    db = sqlite_open(db_name, 0, &err);
    if (db == NULL) {
        fprintf(stderr, "sqlite_open(): %s\n", err);
        free(err);
        return EXIT_FAILURE;
    }

    while (1) {
        /* Get a line of input from the user. */
        printf("> "); fflush(stdout);
        line_len = getline(&line, &line_cap, stdin);
        printf("\n");
        if (line_len == -1) {
            if (!feof(stdin)) perror("getline()");
            break;
        }
        if (line[line_len - 1] != '\n') line[line_len++] = '\n';

        /* Simple-minded scanner. */
        for (i = 0; i < line_len && !isspace(line[i]); ++i);
        if (i == line_len) {cmd = NULL;} else {cmd = line; line[i] = 0;}
        for (++i; i < line_len &&  isspace(line[i]); ++i);
        if (i == line_len) {args = NULL;} else {args = &line[i];}
        line[line_len - 1] = 0;

        if (cmd == NULL) {
            /* Blank line. */
            continue;
        } else if (strcasecmp(cmd, "dict") == 0) {
            /* Dictionary listing. */
            if (sqlite_exec(db, "select type_id, name from dictionary;",
                            dict_cb, NULL, &err) != SQLITE_OK) {
                fprintf(stderr, "sqlite_exec(): %s\n", err);
                free(err);
            }
        } else if (strcasecmp(cmd, "inv") == 0) {
            /* Inventory listing. */
            if (sqlite_exec(db, "  select d.name, count(*)               "
                                "    from dictionary as d, inventory as i"
                                "   where d.type_id = i.type_id          "
                                "group by d.type_id                      "
                                "order by d.name;                        ",
                            inv_cb, NULL, &err) != SQLITE_OK) {
                fprintf(stderr, "sqlite_exec(): %s\n", err);
                free(err);
            }
        } else if (strcasecmp(cmd, "r-list") == 0) {
            /* Recipe listing. */
            if (sqlite_exec(db, r_list_sql, rlist_cb, NULL, &err) !=SQLITE_OK) {
                fprintf(stderr, "sqlite_exec(): %s\n", err);
                free(err);
            }
        } else if (strcasecmp(cmd, "r-info") == 0) {
            /* Recipe details. */
        } else if (strcasecmp(cmd, "add") == 0) {
            /* Add item to shelf. */
            char buf[256];
            snprintf(buf, 256, "insert into inventory values(%u, %u);",
                strtoul(args, NULL, 16), strtoul(args + 9, NULL, 16));
            if (sqlite_exec(db, buf, NULL, NULL, &err) != SQLITE_OK) {
                fprintf(stderr, "sqlite_exec(): %s\n", err);
                free(err);
            }
        } else if (strcasecmp(cmd, "remove") == 0) {
            /* Remove item from shelf. */
        } else if (strcasecmp(cmd, "quit") == 0) {
            /* Exit program. */
            break;
        } else {
            /* Help. */
            printf(
   "  dict          : Display list of all known items.\n"
   "  inv           : Display contents of shelf.\n"
   "  r-list        : Display all possible recipes.\n"
   "  r-info \e[1mRECIPE\e[0m : Give details about \e[1mRECIPE\e[0m.\n"
   "  add \e[1mTAG\e[0m       : Adds item with \e[1mTAG\e[0m to shelf.\n"
   "  remove \e[1mTAG\e[0m    : Removes item with \e[1mTAG\e[0m from shelf.\n"
   "  quit          : Exits the program.\n"
            );
        }
        printf("\n");
    }

    sqlite_close(db);
    free(line);

    return EXIT_SUCCESS;
}

/* vim: set ts=4 sw=4 sts=4 et: */
