/* libgoth/queue.c
 *
 * Byte-oriented ring buffer structure.
 *
 * Copyright (C) 2003, 2004
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* === Header files. ====================================================== */

#include <stdarg.h>
#include <stdlib.h>

#include "libgoth/queue.h"

/* === Non-static function definitions. =================================== */

/* Creates a new empty queue with room for 'size' bytes. */
int queue_new(queue_t** q, int size)
{
    queue_t* qn;
    if (q == NULL || size < 0) return -1;

    qn = malloc(sizeof(*qn));
    if (qn == NULL) return -1;

    if (size > 0) {
        qn->beg = qn->ins = qn->ext = malloc(size + 1);
        if (qn->beg == NULL) return -1;
        qn->end = qn->beg + size + 1;
    } else {
        qn->beg = qn->end = qn->ins = qn->ext = NULL;
    }

    *q = qn;
    return 0;
}

/* Deallocates a queue.  All data stored in the queue is lost. */
int queue_del(queue_t* q)
{
    if (q == NULL) return -1;
    free(q->beg); free(q);
    return 0;
}

/* Resizes a queue to have room for 'new_size' bytes. */
int queue_resize(queue_t* q, int new_size)
{
    int ret;
    if (q == NULL || new_size < 0) return -1;

    /* Check to see how many bytes are in the queue.  On error, bail. */
    ret = queue_count(q);
    if (ret < 0) return ret;
    if (ret != 0 && new_size <= ret) new_size = ret + 1;

    /* There are way too many cases to handle... grumble. */
    if (q->beg == NULL) {
        if (new_size > 0) {
            /* The queue currently has a size of zero.  Allocate. */
            q->beg = malloc(new_size + 1);
            if (q->beg == NULL) return -1;
            q->end = q->beg + new_size + 1;
            q->ins = q->ext = q->beg;
        } else {
            /* Do nothing. */
        }
    } else if (new_size == q->end - q->beg - 1) {
        /* No change in size:  Do nothing. */
    } else if (ret == 0) {
        /* The queue currently has a count of zero. */
        if (new_size == 0) {
            /* Make an empty queue. */
            free(q->beg);
            q->beg = q->end = q->ins = q->ext = NULL;
        } else {
            /* Allocate. */
            char* beg = realloc(q->beg, new_size + 1);
            if (beg == NULL) return -1;
            q->beg = q->ins = q->ext = beg;
            q->end = beg + new_size + 1;
        }
    } else if (q->ins > q->ext) {
        /* Data currently in queue is contiguous. */
        char* beg;
        if (q->beg + new_size + 1 < q->ins) {
            /* Move data to beginning of buffer. */
            memmove(q->beg, q->ext, q->ins - q->ext);
            q->ins += q->beg - q->ext;
            q->ext = q->beg;
        } else if (q->beg + new_size + 1 == q->ins) {
            /* After resize, q->ins would equal q->end.  Fix. */
            q->ins = q->beg;
        }
        beg = realloc(q->beg, new_size + 1);
        if (beg == NULL) return -1;
        q->end = beg + new_size + 1;
        q->ins += beg - q->beg;
        q->ext += beg - q->beg;
        q->beg = beg;
    } else {
        /* Data currently in queue is discontiguous. */
        int tail = q->end - q->ext;
        if (new_size + 1 < q->end - q->beg) {
            /* Shrink. */
            char* beg;
            memmove(q->beg + new_size + 1 - tail, q->ext, tail);
            beg = realloc(q->beg, new_size + 1);
            if (beg == NULL) return -1;
            q->end = beg + new_size + 1;
            q->ins += beg - q->beg;
            q->beg = beg;
        } else {
            /* Grow. */
            char* beg = realloc(q->beg, new_size + 1);
            if (beg == NULL) return -1;
            q->end = beg + new_size + 1;
            q->ins += beg - q->beg;
            q->ext += beg - q->beg;
            q->beg = beg;
            memmove(q->end - tail, q->ext, tail);
        }
        q->ext = q->end - tail;
    }

    return new_size;
}

/* Adds data to a queue, resizing if necessary to store all incoming bytes. */
int queue_push(queue_t* q, char* data, int count)
{
    if (q == NULL || data == NULL || count < 0) return -1;
    if (count == 0) return 0;

    /* Overallocation policy: get MAX(count * 3 / 2, 64) additional bytes. */

    /* WARNING: Reading the following code may damage your sanity. */
    if (q->beg == NULL) {
        /* The queue currently has a size of zero.  Allocate. */
        int size = 1 + (count > 42 ? count * 3 / 2 : 64);
        q->beg = malloc(size);
        if (q->beg == NULL) return -1;
        q->end = q->beg + size;
        q->ins = q->beg + count;
        q->ext = q->beg;
        memcpy(q->beg, data, count);
    } else if (q->ins == q->ext) {
        /* The queue currently has a count of zero. */
        if (count < q->end - q->beg) {
            /* Copy. */
            memcpy(q->beg, data, count);
            q->ins = q->beg + count;
            q->ext = q->beg;
        } else {
            /* Reallocate and copy. */
            int size = q->end - q->beg + (count > 42 ? count * 3 / 2 : 64);
            char* beg = realloc(q->beg, size);
            if (beg == NULL) return -1;
            q->beg = beg;
            q->end = beg + size;
            q->ins = beg + count;
            q->ext = beg;
            memcpy(beg, data, count);
        }
    } else if (q->ins > q->ext) {
        /* Data currently in queue is contiguous. */
        if (q->ins + count < q->end) {
            /* Simply copy. */
            memcpy(q->ins, data, count);
            q->ins += count;
        } else if (q->ins + count == q->end) {
            if (q->ext == q->beg) {
                /* Reallocate.  The queue remains contiguous. */
                int size = q->end - q->beg + (count > 42 ? count * 3 / 2 : 64);
                char* beg = realloc(q->beg, size + 1);
                if (beg == NULL) return -1;
                q->end = beg + size;
                q->ins += beg - q->beg;
                q->ext = beg;
                q->beg = beg;

                memcpy(q->ins, data, count);

                q->ins += count;
            } else {
                /* Copy. */
                memcpy(q->ins, data, count);
                q->ins = q->beg;
            }
        } else if (q->ins - q->ext + count < q->end - q->beg) {
            /* Make queue contiguous. */
            memmove(q->beg, q->ext, q->ins - q->ext);
            memcpy(q->beg + (q->ins - q->ext), data, count);
            q->ins = q->beg + count;
            q->ext = q->beg;
        } else {
            /* Reallocate.  Make queue contiguous. */
            int size = q->end - q->beg + (count > 42 ? count * 3 / 2 : 64);
            char* beg = realloc(q->beg, size);
            if (beg == NULL) return -1;
            q->end = beg + size;
            q->ins += beg - q->beg;
            q->ext += beg - q->beg;
            q->beg = beg;

            memmove(q->beg, q->ext, q->ins - q->ext);
            memcpy(q->beg + (q->ins - q->ext), data, count);

            q->ins = q->beg + count;
            q->ext = q->beg;
        }
    } else {
        /* Data currently in queue is discontiguous. */
        if (q->ins + count < q->ext) {
            /* Copy. */
            memcpy(q->ins, data, count);
            q->ins += count;
        } else {
            /* Reallocate.  queue remains discontiguous. */
            int tail = q->end - q->ext;
            int size = q->end - q->beg + (count > 42 ? count * 3 / 2 : 64);
            char* beg = realloc(q->beg, size);
            if (beg == NULL) return -1;
            q->end = beg + size;
            q->ins += beg - q->beg;
            q->ext += beg - q->beg;
            q->beg = beg;

            memmove(q->end - tail, q->ext, tail);
            memcpy(q->ins, data, count);

            q->ins += count;
            q->ext = q->end - tail;
        }
    }

    return count;
}

/* Moves data from a queue to an outside array. */
int queue_pop(queue_t* q, char* data, int count)
{
    /* Copy from the queue to the output array.  In case of error, or if the
     * queue is empty, return to the caller immediately. */
    int ret = queue_peek(q, data, count);
    if (ret <= 0) return ret;

    /* Discard all bytes successfully copied in the last operation. */
    return queue_discard(q, ret);
}

/* Copies data from a queue to an outside array.  The queue is not modified. */
int queue_peek(queue_t* q, char* data, int count)
{
    int ret;
    if (q == NULL || data == NULL || count < 0) return -1;
    if (count == 0) return 0;

    /* Check to see how many bytes are in the queue_t.  On error, tell the
     * caller.  If the queue_t is empty, simply return 0. */
    ret = queue_count(q);
    if (ret <= 0) return ret;
    if (count > ret) count = ret;

    /* Do the copy. */
    if (q->ext + count <= q->end) {
        /* Joy, contiguous. */
        memcpy(data, q->ext, count);
    } else {
        /* Ick, discontiguous. */
        memcpy(data, q->ext, q->end - q->ext);
        memcpy(data + (q->end - q->ext), q->beg, q->ins - q->beg);
    }

    return count;
}

/* Calls 'cb'() to load data into a queue. */
int queue_fill(queue_t* q, int count, int align, queue_fill_cb_t cb, ...)
{
    int ret, total = 0, goal, avail;
    va_list ap;
    if (q == NULL || func == NULL) return -1;
    if (count == 0) return 0;

    avail = queue_available(q);
    if (avail < 0) return -1;
    if (count < 0 && avail == 0) count = 64;
    if (avail < count) {
        ret = queue_resize(q, queue_size(q) + count);
        if (ret < 0) return ret;
    } else if (count < 0) {
        count = avail;
    }

    while (total < count) {
    }

    if (q->ext < q->ins) {
        /* Free space is discontiguous. */
        goal = q->end - q->ins;
        if (goal > count) {
            /* Do the first of two passes. */
            va_start(ap, cb);
            total = ret = cb(q->ins, goal, ap);
            va_end(ap);
            if (ret < 0) {
                /* Error. */
                return ret;
            } else if (ret > goal) {
                /* Illegal return value. */
                return -1;
            } else if (ret < goal) {
                /* Did not read all bytes.  Quit now. */
                q->ins += ret;
                return ret;
            } else {
                /* cb() read everything.  Wrap pointer and fall through to next
                 * pass. */
                q->ins = q->beg;
            }
        }
    }

    /* Free space is contiguous. */
    if (q->ext == q->ins) goal = q->end - q->ins - 1;
    else                  goal = q->ext - q->ins - 1;
    va_start(ap, func);
    ret = func(q->ins, goal, ap);
    va_end(ap);
    if (ret < E_OK) {
        /* Error. */
        return ret;
    } else if (ret > goal) {
        /* Illegal return value. */
        return -E_INVAL;
    } else {
        /* Success. */
        q->ins += ret;
        if (q->ins == q->end) q->ins = q->beg;
        return count + ret;
    }
}

/* Calls 'func'() to process 'count' bytes taken from the 'data' array, which
 * points inside 'q''s buffer.  Additional arguments are passed through 'ap'. */
int queue_drain(queue_t* q, queue_flow_func_t func, ...)
{
    int ret, count = 0;
    va_list ap;
    if (q == NULL || func == NULL) return -E_INVAL;

    if (queue_empty(q)) {
        /* No bytes to send. */
        count = 0;
    } else if (q->ins < q->ext) {
        /* Data is discontiguous. */
        va_start(ap, func);
        count = ret = func(q->ext, q->end - q->ext, ap);
        va_end(ap);
        if (ret < E_OK) {
            /* Error. */
            return ret;
        } else if (ret > q->end - q->ext) {
            /* Illegal return value. */
            return -E_INVAL;
        } else if (ret < q->end - q->ext) {
            /* Did not send all bytes.  Quit. */
            q->ext += ret;
            return ret;
        } else {
            /* Sent everything.  Wrap pointer and do next block. */
            q->ext = q->beg;
        }
    }

    /* Data is contiguous. */
    va_start(ap, func);
    count += ret = func(q->ext, q->ins - q->ext, ap);
    va_end(ap);
    if (ret < E_OK) {
        /* Error. */
        return ret;
    } else if (ret > q->ins - q->ext) {
        /* Illegal return value. */
        return -E_INVAL;
    } else {
        /* Success. */
        q->ext += ret;
        return count;
    }
}

/* Starting at byte 'pos', scans each byte of queue 'q' using 'func'.  When
 * 'func' returns 1, this function returns the index of the byte for which it
 * did so.  Returns -E_SEARCH if 'func' never returns 1.  If 'pos' is negative,
 * it relative to the end (insertion point) of the queue. */
int queue_scan(queue_t* q, int pos, queue_scan_func_t func, ...)
{
    va_list ap;
    int count, ret;
    char* p, *end;
    if (q == NULL || func == NULL) return -E_INVAL;

    /* Normalize pos. */
    count = queue_count(q);
    if (pos > count)  return -E_SEARCH;
    if (pos < -count) pos = 0;
    else if (pos < 0) pos += count;

    /* Find the start and end pointers. */
    p = q->ext + pos;
    if (p > q->end) p += q->beg - q->end;
    end = (p > q->ins) ? q->end : q->ins;

    while (1) {
        for (; p != end; ++p, ++pos) {
            /* Call func to check this byte. */
            va_start(ap, func);
            ret = func(*p, ap);
            va_end(ap);

            if (ret < E_OK) return ret; /* Error.  :^(     */
            if (ret == 1)   return pos; /* Search success! */
            else                      ; /* Try harder.     */
        }

        /* Contiguous: control reaching this point means search failure. */
        if (end == q->ins) return -E_SEARCH;

        /* Discontiguous: loop a second time. */
        end = q->ins;
    }
}

/* Sets *'data' to point to a linear array of 'count' bytes taken from the
 * queue_t 'q'.  Returns the number of bytes in the array (artificially
 * limited to 'count') on success and an error code on failure. */
int queue_index(queue_t* q, char** data, int count)
{
    int ret;
    if (q == NULL || data == NULL || count < 0) return -E_INVAL;
    if (count == 0) return 0;

    /* Check to see how many bytes are in the queue_t.  On error, tell the
     * caller.  If the queue_t is empty, simply return 0. */
    ret = queue_count(q);
    if (ret <= 0) return ret;
    if (count > ret) count = ret;

    if (q->ext + count <= q->end) {
        /* The data's already contiguous:  Nothing needs to be done. */
    } else {
        /* The data is discontiguous, so realign it. */
        if (q->ext - q->ins > q->end - q->ext) {
            /* A realloc() is necessary to provide enough free space in the
             * buffer to rearrange the data in-place. */
            int tail = q->end - q->ext;
            char* beg = realloc(q->beg, q->ins - q->beg + 2 * tail);
            if (beg == NULL) return -E_MEMORY;
            q->end = beg + (q->ins - q->beg) + 2 * tail;
            q->ins += beg - q->beg;
            q->ext += beg - q->beg;
            q->beg = beg;

            /* Now shuffle the data. */
            memmove(q->end - tail, q->ext, tail);
            memmove(q->beg + tail, q->beg, q->ins - q->beg);
            memmove(q->beg, q->end - tail, tail);
        } else {
            /* There's enough free space in the buffer to rearrange the data
             * in-place. */
            memmove(q->beg + (q->end - q->ext), q->beg, q->ins - q->beg);
            memmove(q->beg, q->ext, q->end - q->ext);
        }

        /* Now the data is one contiguous block starting at q->beg. */
        q->ext = q->beg;
        q->ins = q->ext + ret;
    }

    /* Done!! */
    *data = q->ext;
    return count;
}

/* Throws away up to 'count' bytes from queue_t 'q'.  Returns the number of
 * discarded bytes on success and an error code on failure. */
int queue_discard(queue_t* q, int count)
{
    int ret;
    if (q == NULL || count < 0) return -E_INVAL;
    if (count == 0) return 0;

    /* Check to see how many bytes are in the queue_t.  On error, tell the
     * caller.  If the queue_t is empty, simply return 0. */
    ret = queue_count(q);
    if (ret <= 0) return ret;

    if (count >= ret) {
        /* Empty the entire queue_t by resetting the pointers. */
        q->ins = q->ext = q->beg;
        return ret;
    } else {
        /* It's quite easy to discard data.  Just advance the pointer. */
        q->ext += count;

        /* Wrap the pointer if necessary. */
        if (q->ext >= q->end) q->ext += q->beg - q->end;

        return count;
    }
}

/* Returns the number of bytes currently stored in queue_t 'q'. */
int queue_count(queue_t* q)
{
    if (q == NULL)        return -E_INVAL;
    if (q->beg == NULL)   return 0;
    if (q->ins >= q->ext) return q->ins - q->ext;
                          return (q->end - q->ext) + (q->ins - q->beg);
}

/* Returns the current size in bytes of queue_t 'q'. */
int queue_size(queue_t* q)
{
    if (q == NULL)        return -E_INVAL;
    if (q->beg == NULL)   return 0;
                          return q->end - q->beg - 1;
}

/* Returns true if there are no bytes in 'q'. */
int queue_empty(queue_t* q)
{
    if (q == NULL)        return -E_INVAL;
    if (q->beg == NULL)   return 1;
    if (q->ins == q->ext) return 1;
                          return 0;
}

/* Returns true if writing one more byte to 'q' would require a resize. */
int queue_full(queue_t* q)
{
    if (q == NULL)                                return -E_INVAL;
    if (q->beg == NULL)                           return 1;
    if (q->ins == q->ext - 1)                     return 1;
    if (q->ins == q->end - 1 && q->ext == q->beg) return 1;
                                                  return 0;
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
