/* libgoth/queue.h
 *
 * Byte-oriented ring buffer structure.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

#ifndef SEEN_LIBGOTH_QUEUE_H
#define SEEN_LIBGOTH_QUEUE_H

#include <stdarg.h>

#if 0

/* Summary: */

typedef struct queue {char* beg, *end, *ins, *ext;} queue_t;

typedef int (*queue_fill_cb_t )(char* data, int count, va_list ap);
typedef int (*queue_drain_cb_t)(char* data, int count, va_list ap);
typedef int (*queue_scan_cb_t )(char data, va_list ap);

int queue_new       (queue_t** q, int size);
int queue_del       (queue_t* q);
int queue_resize    (queue_t* q, int new_size);
int queue_push      (queue_t* q, char* data, int count);
int queue_pop       (queue_t* q, char* data, int count);
int queue_peek      (queue_t* q, char* data, int count);
int queue_fill      (queue_t* q, int count, int align, queue_fill_cb_t cb, ...);
int queue_drain     (queue_t* q, int count, int align, queue_drain_cb_t cb,...);
int queue_scan      (queue_t* q, int pos, queue_scan_cb_t cb, ...);
int queue_index     (queue_t* q, char** data, int count);
int queue_discard   (queue_t* q, int count);
int queue_count     (queue_t* q);
int queue_size      (queue_t* q);
int queue_available (queue_t* q);
int queue_empty     (queue_t* q);
int queue_full      (queue_t* q);

#endif

/* Just your average, everyday circular buffer structure...  This structure and
 * its accessor functions are designed to be quite efficient for handling byte-
 * oriented data pipelines.
 *
 * If beg is NULL, then the queue has a size of zero.
 *
 * If ins equals ext, then the queue has a count of zero (is empty).  Because
 * of this, it's necessary to allocate one extra byte, or else there would be
 * no way to differentiate between empty and full.  So, end minus beg equals
 * the queue's size plus one. */
typedef struct queue {
    char* beg;      /* Start of queue's data buffer.   */
    char* end;      /* First byte after end of buffer. */
    char* ins;      /* Location of next insertion.     */
    char* ext;      /* Location of next extraction.    */
} queue_t;

/* Typedef: queue_fill_cb_t
 *   Callback function for use with queue_fill() to load data into a queue.
 *
 * Argument: char* data
 *   Pointer to the start of a byte array to which this function should write.
 *
 * Argument: int count
 *   Maximum number of bytes that this function should write.
 *
 * Argument: va_list ap
 *   Mechanism for passing any additional arguments.  This comes from the "..."
 *   parameter to queue_fill().
 *
 * Return value: int
 *   On success, this function should return the number of bytes it wrote to
 *   'data'.  This value must be no greater than 'count'.  On failure, this
 *   function can return any negative number, which will be passed to the
 *   caller of queue_fill(). */
typedef int (*queue_fill_cb_t)(char* data, int count, va_list ap);

/* Typedef: queue_drain_cb_t
 *   Callback function for use with queue_drain() to process queue data.
 *
 * Argument: char* data
 *   Pointer to the start of a byte array from which this function should read.
 *
 * Argument: int count
 *   Maximum number of bytes available in the 'data' array.
 *
 * Argument: va_list ap
 *   Mechanism for passing any additional arguments.  This comes from the "..."
 *   parameter to queue_drain().
 *
 * Return value: int
 *   On success, this function should return the number of bytes it consumed
 *   in the 'data' array.  This value must be no greater than 'count'.  On
 *   failure, this function can return any negative number, which will be
 *   passed to the caller of queue_drain().  This many bytes of data are
 *   removed from the queue after this function returns.  It is only safe for
 *   this function to read more bytes than it consumes if the 'align' option
 *   of queue_fill() is enabled. */
typedef int (*queue_drain_cb_t)(char* data, int count, va_list ap);

/* Typedef: queue_scan_cb_t
 *   Callback function for use with queue_scan() to search for a data byte
 *   fulfilling certain criteria.  Assuming queue_scan() is being called
 *   correctly, this function is called exactly once for each data byte loaded
 *   into the queue, in order of arrival.
 *
 * Argument: char data
 *   Individual byte to be inspected by this function.
 *
 * Argument: va_list ap
 *   Mechanism for passing any additional arguments.  This comes from the "..."
 *   parameter to queue_scan().
 *
 * Return value: int
 *   If 'data' fulfills the criteria (for example, it's the end of a data
 *   packet), this function should return positive, otherwise zero.  If this
 *   function returns a negative value, queue_scan() aborts and returns that
 *   same value to its caller. */
typedef int (*queue_scan_cb_t)(char data, va_list ap);

/* Function: queue_new()
 *   Creates a new empty queue with room for 'size' bytes.
 *
 * Argument: queue_t** q
 *   The address of a pointer that this function will modify to point to the
 *   new queue structure.
 *
 * Argument: int size
 *   Initial capacity in bytes of the new queue.  It's safe to specify 0 here
 *   because the queue will grow to accomodate the data stored in it.
 *
 * Return value: int
 *   On success, the initial capacity in bytes of the queue.  On failure, -1. */
extern int queue_new(queue_t** q, int size);

/* Function: queue_del()
 *   Deallocates a queue.  All data stored in the queue is lost.
 *
 * Argument: queue_t* q
 *   The queue to be deallocated.
 *
 * Return value: int
 *   On success, 0.  On failure, -1. */
extern int queue_del(queue_t* q);

/* Function: queue_resize()
 *   Resizes a queue to have room for 'new_size' bytes.  Calling this function
 *   is rarely necessary since queues grow automatically, but since they never
 *   shrink on their own queue_resize() might be used to reclaim memory.  If
 *   the data in the queue will not fit in the requested size, the queue will
 *   shrink as much as possible without discarding any data.
 *
 * Argument: queue_t* q
 *   The queue to be resized.
 *
 * Argument: int new_size
 *   Target capacity of the queue.
 *
 * Return value: int
 *   On success, the final size in bytes of the queue.  On failure, -1. */
extern int queue_resize(queue_t* q, int new_size);

/* Function: queue_push()
 *   Adds data to a queue, resizing if necessary to store all incoming bytes.
 *
 * Argument: queue_t* q
 *   Queue in which the data will be stored.
 *
 * Argument: char* data
 *   Pointer to an array of bytes to be added to the queue.
 *
 * Argument: int count
 *   Number of bytes in the 'data' array.
 *
 * Return value: int
 *   On success, the number of bytes written to the queue.  On failure, -1. */
extern int queue_push(queue_t* q, char* data, int count);

/* Function: queue_pop()
 *   Moves data from a queue to an outside array.
 *
 * Argument: queue_t* q
 *   Queue from which data should be removed.
 *
 * Argument: char* data
 *   Pointer to a byte array to which the data will be moved.
 *
 * Argument: int count
 *   Number of bytes to move from the queue to the 'data' array.  If the queue
 *   contains less than 'count' bytes, then all available data is moved,
 *   leaving the queue empty.
 *
 * Return value: int
 *   On success, the number of bytes removed from 'q' and copied to 'data'.  On
 *   failure, -1.  */
extern int queue_pop(queue_t* q, char* data, int count);

/* Function: queue_peek()
 *   Copies data from a queue to an outside array.  The queue is not modified.
 *
 * Argument: queue_t* q
 *   Queue which will serve as the data source.
 *
 * Argument: char* data
 *   Pointer to a byte array which will receive the data copied from 'q'.
 *
 * Argument: int count
 *   Number of bytes to copy from 'q' to 'data'.  If the queue doesn't contain
 *   enough data, all available is copied.
 *
 * Return value: int
 *   On success, the number of bytes copied.  On failure, -1. */
extern int queue_peek(queue_t* q, char* data, int count);

/* Function: queue_fill()
 *   Calls 'cb'() to load data into a queue.
 *
 * Argument: queue_t* q
 *   Queue which will be filled with data.
 *
 * Argument: int count
 *   Maximum number of bytes to load into the queue.  If this is negative, then
 *   queue_fill() attempts to fill the queue to capacity, and if this value is
 *   greater than the queue's capacity, the queue is resized to accomodate.  If
 *   'count' is negative, queue_fill() tries to load a minimum of 64 bytes,
 *   which may mean growing the queue.
 *
 * Argument: int align
 *   If nonzero, 'cb'() is called only once, which may require rearranging the
 *   data already present in the queue to ensure that there are 'count' bytes
 *   of contiguous free space following the queue's insertion pointer.
 *
 * Argument: queue_fill_cb_t cb
 *   Function to call to load data into the queue.  It receives the address and
 *   length of a destination byte array and returns the number of bytes
 *   actually written to the array (which lives inside the queue).  If the free
 *   space in the queue is noncontiguous and the function completely fills the
 *   first block, 'cb'() is called a second time; enable 'align' mode to
 *   prevent this possibility.
 *
 * Argument: ...
 *   Any additional parameters following 'cb' are passed to 'cb'() in the form
 *   of a va_list.  This is useful for providing 'cb' with some context.
 *
 * Return value: int
 *   On success, the number of bytes written to the queue.  On failure, -1, or
 *   the return value of 'cb'() if it returned negative. */
extern int queue_fill(queue_t* q, int count, int align,
                      queue_fill_cb_t cb, ...);

/* Function: queue_drain()
 *   Calls 'cb'() to process data stored in a queue.  This data is then removed
 *   from the queue, hence the term "drain".
 *
 * Argument: queue_t* q
 *   Queue from which the data will be drained.
 *
 * Argument: int count
 *   Maximum number of bytes to drain from the queue.  If this is negative or
 *   greater than the number of bytes in the queue, queue_drain() attempts to
 *   drain all available data.
 *
 * Argument: int align
 *   If nonzero, the data in the queue is aligned in such a way as to call
 *   'cb'() only once.  This is crucial for callback functions which read more
 *   data than they consume in the expectation that the extra data will still
 *   be available the next time queue_drain() is called.
 *
 * Argument: queue_drain_cb_t cb
 *   Function to call to process data in the queue.  It receives the address
 *   and length of a source byte array and returns the number of bytes it
 *   consumed.  Consumed bytes are then removed from the queue, never to be
 *   seen again.  If the data in the queue is noncontiguous and the function
 *   completely consumes the first block, 'cb'() is called a second time, which
 *   in some rare cases can be a serious problem.  Enable 'align' mode to
 *   prevent this possibility.
 *
 * Argument: ...
 *   Any additional parameters following 'cb' are passed to 'cb'() in the form
 *   of a va_list.  This is useful for providing 'cb' with some context.
 *
 * Return value: int
 *   On success, the number of bytes consumed and removed from the queue.  On
 *   failure, -1, or the return value of 'cb'() if it returned negative. */
extern int queue_drain(queue_t* q, int count, int align,
                       queue_drain_cb_t cb, ...);

/* Function: queue_scan()
 *   Calls 'cb'() to search for a byte of data that fulfills some criteria.
 *
 * Argument: queue_t* q
 *   Queue to scan.
 *
 * Argument: int pos
 *   Index of the first byte to scan.  This argument is provided to allow a
 *   scan to pick up where it left off if 'cb'() failed to find a match and
 *   more data arrived.  If negative, 'pos' is taken as being relative to the
 *   end (the insertion point) of the queue, so for instance -1 means the byte
 *   most recently inserted.
 *
 * Argument: int count
 *   The maximum number of bytes to search.  If 'count' is negative or is
 *   greater than or equal to the number of bytes in the queue starting with
 *   index 'pos', then all remaining bytes are searched.
 *
 * Argument: queue_scan_cb_t cb
 *   Function to call for each byte in the queue (starting with the byte at
 *   index 'pos').  It receives each byte passed as a char argument.  If the
 *   byte matches, then 'cb'() should return positive, otherwise zero.  In the
 *   case of error, 'cb'() can return negative to abort the queue_scan().
 *
 * Argument: ...
 *   Any additional parameters following 'cb' are passed to 'cb'() in the form
 *   of a va_list.  This is useful for providing 'cb' with some context.
 *
 * Return value: int
 *   On success, this function returns one plus the index of the first byte for
 *   which 'cb'() returned positive.  This is relative to the extraction point
 *   of the queue, not to 'pos'.  For example, if a queue has five bytes in it,
 *   queue_scan() is called with a 'pos' of -1, and 'cb'() returns positive for
 *   that byte, then queue_scan() returns 5.  If 'cb'() returns positive for
 *   the first byte in the queue (the byte at the extraction point), then
 *   queue_scan() returns 1.  And if 'cb'() always returns zero, queue_scan()
 *   will also return zero to indicate search failure.  Finally, if 'cb'()
 *   returns negative, then queue_scan() aborts and returns that same value.
 *
 * TODO:
 *   It would be far more efficient if this operation could be done in a loop.
 *   But without C++ templates this would mean using #define's.  I do not want
 *   to go that route.
 *
 *   But it may be better still to eliminate or deprecate this function in
 *   favor of an enhanced queue_index() capable of returning a pointer to a
 *   position other than the extraction point. */
extern int queue_scan(queue_t* q, int pos, int count, queue_scan_cb_t cb, ...);

/* Function: queue_index()
 *   Retreives a pointer to a linear array of data contained in the queue.
 *
 * Argument: char** data
 *   Address of a pointer modified by this function to point to the data.  Note
 *   that this pointer may be invalidated by queue operations.
 *
 * Argument: int count
 *   Minimum number of bytes to be made contiguous.  If negative or greater
 *   than or equal to the number of bytes in the queue, the entire queue is
 *   aligned.
 *
 * Return value: int
 *   On success, returns the number of contiguous data bytes available.  If
 *   'count' is negative or greater than or equal to the number of bytes in the
 *   queue, the return value will equal the number of bytes in the queue.  In
 *   other cases, the return value may exceed the number of bytes requested if
 *   more data happens to be already aligned.  On failure, -1 is returned.
 *
 * TODO:
 *   Implement a 'pos' argument, as described in the comment for queue_scan().
 *   I probably should have done this a long time ago, but alas I settled for
 *   an inferior, easier-to-code solution, and now I have to backtrack. */
extern int queue_index(queue_t* q, char** data, int count);

/* Function: queue_discard()
 *   Removes bytes from the queue.
 *
 * Argument: queue_t* q
 *   Queue from which to remove data.
 *
 * Argument: int count
 *   Number of bytes of data to remove from the queue.  If 'count' is
 *   negative or is greater than or equal to the number of bytes in the
 *   queue, it is emptied completely.
 *
 * Return value: int
 *   On success, returns the number of bytes discarded.  On failure, -1. */
extern int queue_discard(queue_t* q, int count);

/* Function: queue_count()
 *   Checks the amount of data in the queue.
 *
 * Argument: queue_t* q
 *   The queue to examine.
 *
 * Return value: int
 *   Returns the number of bytes currently stored in the queue.  On failure,
 *   returns -1. */
extern int queue_count(queue_t* q);

/* Function: queue_size()
 *   Checks the current capacity of the queue.
 *
 * Argument: queue_t* q
 *   The queue to examine.
 *
 * Return value: int
 *   Returns the current number of bytes that can be stored in the queue
 *   without resizing it.  On failure, returns -1. */
extern int queue_size(queue_t* q);

/* Function: queue_available()
 *   Checks the current amount of available free space in the queue.
 *
 * Argument: queue_t* q
 *   The queue to examine.
 *
 * Return value: int
 *   Returns the current number of bytes that can be added to the queue without
 *   resizing it.  On failure, returns -1. */
extern int queue_available(queue_t* q);

/* Function: queue_empty()
 *   Checks the queue to see if it contains any data.
 *
 * Argument: queue_t* q
 *   The queue to examine.
 *
 * Return value: int
 *   Returns 1 if the queue contains at least one byte of data, otherwise 0.
 *   On failure, returns -1. */
extern int queue_empty(queue_t* q);

/* Function: queue_full()
 *   Checks the queue to see if it's full.
 *
 * Argument: queue_t* q
 *   The queue to examine.
 *
 * Return value: int
 *   Returns 1 if writing any more data to the queue would require a resize,
 *   otherwise 0.  On failure, -1. */
extern int queue_full(queue_t* q);

#endif

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
