/* dev/pressure.c
 *
 * Pressure/door sensor.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* TODO:
 *
 * > Complete the calibration code.  Right now the calibration is hard-coded.
 *   The client will "drive" calibration by telling the pressure sensors what
 *   mass is currently placed on which sensor.  Also there needs to be a way to
 *   save the calibration and reload it on system reboot.
 *
 *   I'm thinking that the client should by default do three calibration
 *   samples: empty shelf, medium weight, and maximum weight.  But what's the
 *   maximum weight?  Would it be prohibitive to distribute precision-crafted
 *   weights of that size with the smartshelf?  There should also be an option
 *   to do arbitrary calibration samples, to cope with sensors with hideously
 *   nonlinear responses.
 *
 * > Communication with the pressure sensor should be done through argusd.  I
 *   see three possibilities:
 *
 *     1. argusd uses a shared memory segment.  Then com/argusd.c will need to
 *        register a regular timer alarm upon which it checks the shm for data
 *        and passes it to this code.
 *
 *     2. argusd uses a shm plus a pipe or UNIX socket.  This makes the timer
 *        alarm no longer necessary, since we'll be notified of updates.
 *
 *     3. argusd uses a pipe or UNIX socket.  com/argusd.c (possibly renamed to
 *        com/unix.c or similar) reads data from there and again hands it to
 *        dev/pressure.c.
 *
 *   I'm glad this code is smart enough to cover all these possibilities. :^)
 *
 * > Compensate for jitter.  A simple low-pass filter should work.  The amount
 *   of jitter for each sensor might be detected at calibration time, and this 
 *   value would be used to calculate the quality factor of the filter.
 *
 * > Rememeber the old sensor values, and only report changes.  For this to be
 *   helpful, the jitter correction must work correctly. */

/* === Header files. ====================================================== */

#include <stdio.h>
#include <stdlib.h>

#include "dev/pressure.h"
#include "lib/device.h"
#include "lib/error.h"
#include "lib/misc.h"
#include "lib/vector.h"

/* === Type definitions, static variables, and static prototypes. ========= */

/* A calibration sample. */
typedef struct calib {
    int val;                        /* Value read from sensor.              */
    int mass;                       /* Mass (grams) placed on that sensor.  */
} calib_t;

/* State information for a pressure sensor device.  The current device consists
 * of four pressure sensors and two door sensors. */
typedef struct pressure {
    vector_t* calibs[4];            /* Calibration information.  <calib_t>  */

    int door_status[2];             /* True if each door is opened.         */

    int have_sync;                  /* Currently in sync with device?       */
    int resync_pos;                 /* Number of resync bytes found.        */
    int packet_rem;                 /* Bytes remaining until end of packet. */
    int packet_len;                 /* Number of bytes in a good packet.    */
    int packet_bad;                 /* If true, current packet is bad.      */

#if 0
    shelf_t* shelf;                 /* Associated shelf. */
#endif
} pressure_t;

/* Picks the good bits out of the packet. */
static int pressure_on_packet(device_t* d, char* data, int len);

/* Returns true when the packet delimeter is encountered. */
static int pressure_is_delim(device_t* d, char data);

/* Cleans up the pressure sensor-specific portion of device 'd'. */
static int pressure_on_cleanup(device_t* d);

/* === Non-static function definitions. =================================== */

/* Connects 'd' to a pressure sensor. */
int pressure_open(device_t* d)
{
    int ret, i;
    pressure_t* p;
    if (d == NULL) return -E_INVAL;

    p = malloc(sizeof(*p));
    if (p == NULL) return -E_MEMORY;

    /* Connect to the pressure-specific handlers. */
    d->on_packet      = pressure_on_packet;
    d->is_delim       = pressure_is_delim;
    d->on_cleanup_dev = pressure_on_cleanup;
    d->dev_state      = p;

    /* Start with fake calibration assuming a perfectly linear response within
     * a given range. */
    for (i = 0; i < 4; ++i) {
        calib_t calibs[] = {{0, 0}, {20, 0}, {220, 3000}, {255, 3000}};
        vector_new(&p->calibs[i], LEN(calibs), sizeof(*calibs));
        vector_insert(p->calibs[i], 0, calibs, LEN(calibs));
    }

    /* Other initial settings. */
    p->door_status[0] = 0; p->door_status[1] = 0; p->have_sync  =  0;
    p->resync_pos     = 0; p->packet_rem     = 0; p->packet_len = 13;
    p->packet_bad     = 1;

    return E_OK;
}

/* === Static function definitions. ======================================== */

/* Picks the good bits out of the packet. */
static int pressure_on_packet(device_t* d, char* data, int len)
{
    int sensors[4];
    int i, j;
    pressure_t* p = d->com_state;

    /* Exit early if the current packet is marked as bad. */
    if (p->packet_bad) return;

    /* The first four bytes represent the four pressure sensors. */
    for (i = 0; i < 4; ++i) {
        calib_t* calibs;
        int count = vector_index(p->calibs[i], 0, &calibs, 0);
        int x = data[i];

        for (j = 0; j < count - 1; ++j) {
            int l = calibs[i + 0].val;      /* "Left".   */
            int r = calibs[i + 1].val;      /* "Right".  */
            int b = calibs[i + 0].mass;     /* "Bottom". */
            int t = calibs[i + 1].mass;     /* "Top".    */
            if (l <= x && x <= r) {
                /* This is the appropriate calibration sample. */
                sensors[i] = b + (x - l) * (t - b) / (r - l);
                break;
            }
        }
    }

    /* The two most-significant bits of the ninth byte are the door sensors. */
    for (i = 0; i < 2; ++i) {
        int bit = 1 << (7 - i);
        if ((p->door_status[i] ? bit : 0) != (data[8] & bit)) {
            /* The door opened or closed. */
            p->door_status[i] = !p->door_status[i];
        }
    }

    /* TODO: inform high-level code of data changes. */

    return E_OK;
}

/* Returns true when the packet delimeter is encountered. */
static int pressure_is_delim(device_t* d, char data)
{
    pressure_t* p = d->com_state;

    if (p->have_sync) {
        /* So far, we're on track.  Count down to the end of the packet. */
        --p->packet_rem;

        /* Check the three trailer bytes. */
        if (    (p->packet_rem == 2 && data != '\r') ||
                (p->packet_rem == 1 && data != '\n') ||
                (p->packet_rem == 0 && data != '>' )) {
            /* Aargh, we didn't get the expected trailer byte.  Resync. */
            p->have_sync  = 0;
            p->resync_pos = 0;
            p->packet_bad = 1;
            return 0;
        }

        if (p->packet_rem == 0) {
            /* Aha, the packet end! */
            p->packet_bad = 0;
            p->packet_rem = p->packet_len;
            return 1;
        } else {
            /* Not yet the end of the packet... */
            return 0;
        }
    } else {
        /* We don't have synchronization, so attempt to resync by treating the
         * next "\r\n>" we see as the end of the packet. */
        switch (p->resync_pos) {
        case 0:
            if (data == '\r') {
                /* Cool, possibly the first byte of the framing marker. */
                p->resync_pos = 1;
            }
            return 0;
        case 1:
            if (data == '\n') {
                /* Possibly the second byte! */
                p->resync_pos = 2;
            } else {
                /* Drat, the preceding '\r' wasn't a framing marker. */
                p->resync_pos = 0;
            }
            return 0;
        case 2:
            if (data == '>' ) {
                /* Yay, resync!  But since the buffer might not contain the
                 * entire packet (which was probably corrupted anyway), we
                 * can't really use it, so don't touch the packet_bad flag
                 * (which at this point is set).  With luck, the next packet
                 * will be good. */
                p->have_sync = 1;
                p->packet_rem = p->packet_len;
                return 1;
            } else {
                /* Got '\r' and '\n', but they weren't framing markers. */
                p->resync_pos = 0;
                return 0;
            }
        }
    }
}

/* Cleans up the pressure sensor-specific portion of device 'd'. */
static int pressure_on_cleanup(device_t* d)
{
    int i;
    pressure_t* p = d->dev_state;
    for (i = 0; i < 4; ++i)
        vector_del(p->calibs[i]);
    free(p);
    return E_OK;
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
