/* lib/error.c
 *
 * Error handling junk.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* === Header files. ====================================================== */

#include <errno.h>
#include <printf.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lib/error.h"

/* === Type definitions, static variables, and static prototypes. ========= */

/* Name of the program. */
static char* error_prog_name;

/* Prints the current error. */
static int printf_error(FILE*, const struct printf_info*, const void* const*);

/* Gives information on the arguments taken by %!. */
static int printf_error_arginfo(const struct printf_info*, size_t n, int*);

/* Helper function for error_print() and error_bail(). */
static void error_vprint(int e, const char* func, const char* mesg, va_list ap);

/* === Non-static variable definitions. ==================================== */

/* List of short error names. */
const char* error_name[] = {
"E_OK", "E_ERROR", "E_ERRNO", "E_INVAL", "E_MEMORY", "E_SEARCH", "E_ASSERT"
};

/* List of short error descriptions. */
const char* error_desc[] = {
/* E_OK     */  "Success.",
/* E_ERROR  */  "Generic error.",
/* E_ERRNO  */  "libc error \"%!\".",
/* E_INVAL  */  "Bad param or param out of range.",
/* E_MEMORY */  "Out of memory.",
/* E_SEARCH */  "Search failure.",
/* E_ASSERT */  "Error in program.",
};

/* === Non-static function definitions. =================================== */

/* Initializes error printing.  Sets the program name to 'prog_name'. */
int error_init(const char* prog_name)
{
    /* Remember the name of the program. */
    error_prog_name = strdup(prog_name);

    /* Set up error printing. */
    if (register_printf_function('!', printf_error,
            printf_error_arginfo) == -1) {
        error_print(-E_ERROR, "error_init()", "Can't register \"%%!\" "
                    "printf handler.");
    }
    return E_OK;
}

/* Prints the error associated with 'e'.  Also prints 'func' if supplied.  If
 * 'mesg' is non-NULL, it is printed in place of 'e'. */
int error_print(int e, const char* func, const char* mesg, ...)
{
    va_list ap;
    va_start(ap, mesg);
    error_vprint(e, func, mesg, ap);
    va_end(ap);
    return E_OK;
}

/* Calls error_print() and terminates with an error code of abs('e'). */
int error_bail(int e, const char* func, const char* mesg, ...)
{
    va_list ap;
    va_start(ap, mesg);
    error_vprint(e, func, mesg, ap);
    va_end(ap);
    exit(abs(e));
}

/* === Static function definitions. ======================================== */

/* Prints the current error. */
static int printf_error(FILE* stream, const struct printf_info* info,
                       const void* const* args)
{
    int len = info->left ? -info->width : info->width;
    return fprintf(stream, "%*s", len, strerror(errno));
}

/* Gives information on the arguments taken by %!. */
static int printf_error_arginfo(const struct printf_info* info,
                               size_t n, int* argtypes)
{
    return 0;
}

/* Helper function for error_print() and error_bail(). */
static void error_vprint(int e, const char* func, const char* mesg, va_list ap)
{
    e = abs(e);
    if (error_prog_name != NULL) fprintf(stderr, "%s: ", error_prog_name);
    if (func != NULL)            fprintf(stderr, "%s: ", func);
    if (mesg == NULL) {
        if (e >= E_USER)         fprintf(stderr, "Custom error #%d.", e-E_USER);
        else if (e >= E_COUNT)   fprintf(stderr, "Undefined error #%d.",e);
        else                     fprintf(stderr, error_desc[e]);
    } else                       vfprintf(stderr, mesg, ap);
    fprintf(stderr, "\n");
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
